# -*- coding: utf-8 -*-
require_dependency 'user'

module SidfmPatches
  module UserPatch
    extend ActiveSupport::Concern

    included do
      unloadable
      has_and_belongs_to_many :sidfmd_categories
      has_and_belongs_to_many :sidfmd_cat_groups
      has_many :sidfmd_cat_groups_users
      has_many :sidfmd_rule_a_mail_categories, dependent: :destroy, foreign_key: 'from_user_id'
      has_many :sidfmd_rule_a_mail_users, dependent: :destroy
      has_many :sidfmd_rule_chp_user_subs, dependent: :destroy
      has_many :sidfmd_pvt_filters, dependent: :destroy
      has_many :sidfmd_vm_apikeys, dependent: :destroy
      has_one :oic_session, dependent: :destroy     

      #パスワードポリシー
      validate do
        errors.add(:password) if !check_password_with_password_policy
      end

      OPENID_CONNECT_AUTH_NAME = 'OpenIDConnect'
      LASTNAME_LENGTH_LIMIT = 30
      FIRSTNAME_LENGTH_LIMIT = 30
      ROLE_SIDFM_ADM = 'SIDfmVM管理者'.freeze
      ROLE_SIDFM_MNG = 'SIDfmVM報告者'.freeze
      ROLE_SIDFM_USE = 'SIDfmVM利用者'.freeze
      ROLE_SIDFM_USR = 'SIDfmVM閲覧者'.freeze
      ROLES_SIDFM = [ ROLE_SIDFM_ADM, ROLE_SIDFM_MNG, ROLE_SIDFM_USE, ROLE_SIDFM_USR ].freeze

      # ユーザの全カテゴリ
      def categories
        SidfmdCategory.where("
          id IN (?) OR id IN (?)
        ",
          self.sidfmd_categories.select(:id),
          SidfmdCatGroup.joins(:sidfmd_categories)
            .where(:id => self.sidfmd_cat_groups_users.select(:sidfmd_cat_group_id))
            .select('sidfmd_categories.id')
        )
      end

      # ユーザのカテゴリグループのカテゴリ
      def cat_group_categories
        SidfmdCategory.where(
          :id => self.sidfmd_cat_groups.joins(:sidfmd_categories).select('sidfmd_categories.id')
        )
      end

      # ユーザのカテゴリグループを更新
      def update_cat_groups(current_user_id:, role_admin:, cat_group_ids:)
        cat_group_ids = cat_group_ids.map(&:to_i)
        if role_admin then
          self.sidfmd_cat_groups = SidfmdCatGroup.where(:id => cat_group_ids)
        else
          #管理者でない場合は、ユーザのカテゴリグループの範囲内のみ更新
          current_user = User.find_by_id(current_user_id)
          if current_user then
            current_user_cat_group_ids = current_user.sidfmd_cat_groups.pluck(:id)
            self.sidfmd_cat_groups_users
                .where(:sidfmd_cat_group_id => current_user_cat_group_ids)
                .destroy_all
            SidfmdCatGroupsUser.import((current_user_cat_group_ids & cat_group_ids).map{|grp_id|
              SidfmdCatGroupsUser.new(user_id: self.id, sidfmd_cat_group_id: grp_id)
            })
          end
        end
      end

      #カテゴリグループのカテゴリを含む、指定したカテゴリに属する全ユーザを取得する
      def self.get_category_users(category_ids)
        self.where("
          id IN (?) OR id IN (?)
        ",
          SidfmdCategoriesUser.where(:sidfmd_category_id => category_ids).select(:user_id),
          SidfmdCatGroupsUser.where(:sidfmd_cat_group_id =>
            SidfmdCatGroupsCategory.where(:sidfmd_category_id => category_ids).select(:sidfmd_cat_group_id)
          ).select(:user_id)  
        )
      end

      # 指定したユーザのカテゴリ・カテゴリグループ変更後の全カテゴリのidを返す
      # カテゴリ・カテゴリグループを指定しない場合は、変更しないものとして扱う
      def self.after_cat_group_change_categories(user_id:, category_ids: nil, cat_group_ids: nil)
        if !category_ids && !cat_group_ids then
          user = User.find_by_id(user_id)
          if user then
            return user.categories.pluck(:id)
          else
            return []
          end
        elsif category_ids && cat_group_ids then
          return SidfmdCategory.where("
            id IN (?) OR id IN (?)
          ",
            category_ids,
            SidfmdCatGroupsCategory.where(:sidfmd_cat_group_id => cat_group_ids).select(:sidfmd_category_id)
          ).pluck(:id)
        elsif category_ids then
          return SidfmdCategory.where("
            id IN (?) OR id IN (?)
          ",
            category_ids,
            SidfmdCatGroupsCategory.where(:sidfmd_cat_group_id =>
              SidfmdCatGroupsUser.where(:user_id => user_id).select(:sidfmd_cat_group_id)
            ).select(:sidfmd_category_id)
          ).pluck(:id)
        else
          return SidfmdCategory.where("
            id IN (?) OR id IN (?)
          ",
            SidfmdCategoriesUser.where(:user_id => user_id).select(:sidfmd_category_id),
            SidfmdCatGroupsCategory.where(:sidfmd_cat_group_id => cat_group_ids).select(:sidfmd_category_id)
          ).pluck(:id)
        end
      end

      # 指定したカテゴリ(id)のみのユーザのidを返す
      # 今のところ、カテゴリ削除時に、カテゴリがなくなるユーザを取得するために使用
      def self.only_category_user_ids(category_ids)
        users_category_ids = {}
        related_user_ids = (
          SidfmdCategoriesUser.where(
            :user_id => SidfmdCategoriesUser.where(:sidfmd_category_id => category_ids).select(:user_id)
          ).pluck(:user_id) +
          SidfmdCatGroupsUser.joins(:sidfmd_cat_group).joins("
            LEFT OUTER JOIN sidfmd_cat_groups_categories ON sidfmd_cat_groups.id = sidfmd_cat_groups_categories.sidfmd_cat_group_id
          ")
          .where(
            :sidfmd_cat_groups_users => {
              :user_id => SidfmdCatGroupsUser.where(
                :sidfmd_cat_group_id => SidfmdCatGroupsCategory.where(:sidfmd_category_id => category_ids).select(:sidfmd_cat_group_id)
              ).select(:user_id)
            }
          ).pluck(:user_id) #カテゴリが設定されているグループのユーザのデータのみを取得
        ).uniq

        SidfmdCategoriesUser.where(:user_id => related_user_ids).each do |cat_user|
          users_category_ids[cat_user.user_id] = [] unless users_category_ids[cat_user.user_id]
          users_category_ids[cat_user.user_id] << cat_user.sidfmd_category_id
        end
        SidfmdCatGroupsUser.joins(:sidfmd_cat_group).joins("
          LEFT OUTER JOIN sidfmd_cat_groups_categories ON sidfmd_cat_groups.id = sidfmd_cat_groups_categories.sidfmd_cat_group_id
        ")
        .where(:sidfmd_cat_groups_users => { :user_id => related_user_ids })
        .select(
          'sidfmd_cat_groups_categories.sidfmd_category_id',
           :user_id
        ).each do |grp_usr|
          users_category_ids[grp_usr.user_id] = [] unless users_category_ids[grp_usr.user_id]
          users_category_ids[grp_usr.user_id] << grp_usr.sidfmd_category_id
        end
        target_user_ids = Set.new
        users_category_ids.each do |user_id, cat_ids|
          target_user_ids.add(user_id) if (cat_ids.compact - category_ids).size == 0
        end
        target_user_ids = target_user_ids.to_a
      end

      # ユーザ一覧
      def self.create_all_content(sidfm_roles:, current_user_id:, user_category_ids:, sidfm_project_id:)
        content = self.where.not(:type => 'AnonymousUser')
        if !sidfm_roles.include?(:admin) then
          # 管理者以外は制限あり
          if sidfm_roles.include?(:user_admin) && user_category_ids.size > 0 then
            userids = self.get_category_users(user_category_ids).pluck(:id)
            content = content.where.not(
              :id => content.joins("
                INNER JOIN members ON members.project_id = #{sidfm_project_id} AND members.user_id = users.id
              ")
              .joins("INNER JOIN member_roles ON member_roles.member_id = members.id")
              .joins("INNER JOIN roles ON roles.id = member_roles.role_id")
              .where("roles.name = ?", "SIDfmVM管理者")
              .select(:id)
            )
            .where(:admin => false)
            .where(:id => userids)
          else
            content = content.where(:id => current_user_id)
          end
        end
        content
      end

      # ユーザ関連のルールを処理
      def self.process_related_rules(**opts)
        r = true
        param = {
          :role_admin => opts[:sidfm_roles].include?(:admin),
          :user_category_ids => (opts[:sidfm_roles].include?(:admin) ? nil : User.current.categories.pluck(:id)),
          :action => opts[:action],
          :messages_necessary => false
        }
        case opts[:action]
        when 'delete' then
          param[:delete_user_ids] = opts[:delete_user_ids]
        when 'category_change', 'bulk_category_change' then
          param[:change_category_user_ids] = opts[:change_category_user_ids]
        when 'auth_change', 'bulk_auth_change' then
          param[:change_role_user_ids] = opts[:change_role_user_ids]
        when 'import' then
          param[:change_category_user_ids] = opts[:change_category_user_ids]
          param[:delete_user_ids] = opts[:delete_user_ids]
          param[:change_role_user_ids] = opts[:change_role_user_ids]
        else
          #invalid
        end
        ret = SidfmdRule.users_related_rule_info(param)
        if ret[:is_unchangeable_user] then
          r = false
        else
          SidfmdRuleChpUserSub.where(:id => ret[:delete_chp_user_sub_ids]).destroy_all if ret[:delete_chp_user_sub_ids].any?
          SidfmdRuleCond.where(:id => ret[:delete_cond_ids]).destroy_all if ret[:delete_cond_ids].any?
          SidfmdRuleAction.where(:id => ret[:delete_action_ids]).destroy_all if ret[:delete_action_ids].any?
          SidfmdRule.where(:id => ret[:delete_rule_ids]).destroy_all if ret[:delete_rule_ids].any?
        end
        r
      end

      #controller側でも使用する
      def validate_password_min_length(password)
        r = true
        min_length = Setting.password_min_length.to_i
        r = false if password.length < min_length
        return r
      end

      def validate_password_symbols(password) 
        r = true
        available_symbols = Sidfm.get_password_available_symbols
        password.gsub(/[a-zA-Z]|[0-9]/, '').each_char do |c|
          if !available_symbols.include?(c) then
            r = false
            break
          end
        end
        return r
      end

      def is_chars_include?(password, regexp)
        password_len = password.length
        #check_password = password.gsub(/#{regexp}/, '')
        check_password = password.gsub(Regexp.new(regexp), '')
        if password_len == check_password.length then #文字数が同じ == 文字が含まれていない
          return false
        else
          return true
        end
      end

      def validate_password_char_types(password)
        r = true
        available_char_types = Sidfm.get_password_available_char_types
        available_symbols = Sidfm.get_password_available_symbols
        is_chars_mix_needed = Sidfm.get_is_password_char_type_mix_needed
        available_char_types = available_char_types.split('/')
        is_en = available_char_types.include?('en')
        is_num = available_char_types.include?('num')
        is_sym = available_char_types.include?('sym')
        regxp_symbols = available_symbols.split('').map{|sym| 
          if "\$()-^|[]+*{}.?/".include?(sym) then
            '\\' + sym
          else
            sym
          end
        }.join
        regxp_symbols = '[' + regxp_symbols + ']'
        #使用許可されていない文字が入っているかチェック
        if !is_en then
          is_chars_included = is_chars_include?(password, '[a-zA-Z]')
          r = false if is_chars_included
        end
        if !is_num then
          is_chars_included = is_chars_include?(password, '[0-9]')
          r = false if is_chars_included
        end
        if !is_sym && available_symbols.size > 0 then
          is_chars_included = is_chars_include?(password, regxp_symbols)
          r = false if is_chars_included
        end
        #文字種の混在チェック
        if is_chars_mix_needed then
          if is_en then
            is_chars_included = is_chars_include?(password, '[a-zA-Z]')
            r = false if !is_chars_included
          end

          if is_num then
            is_chars_included = is_chars_include?(password, '[0-9]')
            r = false if !is_chars_included
          end
           
          if is_sym && available_symbols.size > 0 then
            is_chars_included = is_chars_include?(password, regxp_symbols)
            r = false if !is_chars_included
          end
        else
          check_password = password
          check_password = check_password.gsub(/[a-zA-Z]/, '') if is_en
          check_password = check_password.gsub(/[0-9]/, '') if is_num
          check_password = check_password.gsub(Regexp.new(regxp_symbols), '') if is_sym
          r = false unless check_password.length == 0 #全ての文字が変換されない = 使用不可の文字が含まれている
        end
        return r
      end

      def validate_password_same_as_username(password, username)
        r = true
        is_permited_password_as_same_username = Sidfm.get_is_permited_password_as_same_username
        unless is_permited_password_as_same_username then
          if password == username then
            r = false
          end
        end  
        return r
      end
    
      def validate_password_includes_username(password, username)
        r = true
        is_permited_password_includes_username = Sidfm.get_is_permited_password_includes_username
        unless is_permited_password_includes_username then
          if password.include?(username) then
            r = false
          end
        end
        return r 
      end

      def check_password_with_password_policy
        if User.current.admin == true || !self.password then
          return true
        end
        #文字数のvalidation
        is_pass = validate_password_min_length self.password
        return nil unless is_pass
        #利用可能な記号のvalidation
        is_pass = validate_password_symbols self.password
        return nil unless is_pass  
        #利用可能な文字種のvalidation
        is_pass = validate_password_char_types self.password
        return nil unless is_pass
        if self.login then
          is_pass = validate_password_same_as_username self.password, self.login
          return nil unless is_pass
          is_pass = validate_password_includes_username self.password, self.login
          return nil unless is_pass
        end
        return true
      end
    end
  end
end

ActionDispatch::Reloader.to_prepare do
  unless User.included_modules.include?(SidfmPatches::UserPatch)
    User.send(:include, SidfmPatches::UserPatch)
  end
end
