# -*- coding: utf-8 -*-
require 'date'
require 'csv'
require 'sidfm_controller_support'
require 'sidfm_vulnerability_controller_support'

class SidfmVulnerabilityController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport, SidfmVulnerabilityControllerSupport
  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles
  before_action :get_id

  before_action :get_view_state
  after_action :put_view_state

  before_action :get_vm_version
  before_action :clear_import_session
  before_action :clear_export_session, only: [ :index ]
  before_action -> { get_summary_table_lines('vulnerability') }

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_dates', 'send_mail' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           User.current.categories.order(:id).pluck(:id)
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
    @cache_prefix = @role_admin ? '/vul/admin/0' : "/vul/user/#{@user_category_ids.join(",")}"
  end

  # idを取得する。また、アクセスしているユーザがidに対応する脆弱性を参照可能か確認する。
  # FIXME: 複数のissueが入ることを考慮すること。
  def get_id
    @id = params[:id].to_i
    if @id == 0 and /\AKB(\d+)\z/ =~ params[:id] then
      @id = -Regexp.last_match(1).to_i
    end
    raw_issue = SidfmdIssue.enable_hosts.categorized.where(:html_id => @id)

    issue = raw_issue.count > 0 ? raw_issue.first : nil
    view_issue = SidfmdVulView.find_by(:html_id => @id) 
    @html = {}
    if issue and view_issue then
      @html[:html_id] = issue.html_id
      @html[:title] = SidfmdVulView.display_title(view_issue.title) #issue.title
      #@html[:cve_ids] = issue.sidfmd_issue_cves.order(:cve_id).pluck(:cve_id)
      cvss_v2_vectors = SidfmdIssueCve.get_cvss_v2_vectors
      cvss_v3_vectors = SidfmdIssueCve.get_cvss_v3_vectors
      create_display_vector = ->(vector, target_vector){
        display_vectors = []
        vector.split('/').each do |item|
          item = item.split(':')
          target = target_vector == 'c2_vector' ? cvss_v2_vectors[item[0]] : cvss_v3_vectors[item[0]]
          display_vectors << "#{target['item_name']}: #{target['values'][item[1]]}"
        end
        display_vectors.join('<br>')
      }
      @html[:cves] = []
      issue.sidfmd_issue_cves.order(:cve_id).select(:cve_id, :c2_vector, :c3_vector).each do |cve|
        cve_data = {}
        cve_data[:cve_id] = cve.cve_id
        if cve.c2_vector then
          cve_data[:c2_vector] = create_display_vector.(cve.c2_vector, 'c2_vector')
        end
        if cve.c3_vector then
          cve_data[:c3_vector] = create_display_vector.(cve.c3_vector, 'c3_vector')
        end
        @html[:cves] << cve_data
      end
    else
      redirect_to sidfm_vulnerabilities_path, :notice => '指定された脆弱性IDは存在しません。' and return
    end
    if !@role_admin then
      flg_permit = false
      raw_issue.each do |i|
        if @user_host_ids.include?(i[:sidfmd_host_id])
          flg_permit = true
          break
        end
      end
      unless flg_permit then
        logger.error("Access Denied: user_id=#{User.current.id}, html_id=#{@id}, host_id=#{raw_issue.first[:sidfmd_host_id]}, granted_hosts=#{@user_host_ids.join(',')}")
        redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('対象脆弱性に対するアクセスを許可されていません。') and return
      end
    end
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # カテゴリ選択のみ、反映させる
  def create_all_content(category_id = 0)
    @hosts[:all_content] = SidfmdIssue.create_content(:category_id => category_id, :role_admin => @role_admin, :user_host_ids => @user_host_ids, :user_category_ids => @user_category_ids, :html_id => @id, :corresp_required => @corresp_required)
  end

  def create_host_view_content(category_id = 0)
    @vul_view = SidfmdVulView.find_by(:html_id => @id)
    if @vul_view then
      @hosts[:all_content] = SidfmdHostView.create_vul_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :vul_view_id => @vul_view.id,
        :category_id => category_id
      )
    else
      @hosts[:all_content] = SidfmdHostView.where(:id => -1)
    end
    @all_host_ids = @hosts[:all_content].pluck(:sidfmd_host_id)
  end

  def init_filter_keys
    @filter_keys = {}

    @dist_versions = vul_dist_versions(@hosts[:all_content])
    @filter_keys = default_vul_filters(@dist_versions)

    @default_filters = {}
    @filter_keys.each do |key, value|
      @default_filters[key] = @filter_keys[key][:default]
    end
  end

  def check_filters
    init_filter_keys
    if @hosts[:filters][:categories] && @hosts[:filters][:categories][:ids] then
      @hosts[:filters][:categories][:ids] = @hosts[:filters][:categories][:ids].map(&:to_i)
    end
    @hosts[:filters] = establish_filters_consistency(@hosts[:filters], @filter_keys)
    @hosts[:filters] = filter_filters(
      :filters => @hosts[:filters] ,
      :user_host_osnames => @dist_versions,
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids
    )
    # 脆弱性別状況画面から遷移してきたときは、検索の意味合いが異なるためリセット
    if @filter_keys[:products][:default].keys.size != @hosts[:filters][:products].keys.size then
      @hosts[:filters][:products] = @filter_keys[:products][:default]
    end
    #products = {}
    #@filter_keys[:products][:default].each do |key, val| #脆弱性別状況画面とパラメータが異なるため、修正する
    #  products[key] = @hosts[:filters][:products][key] ? @hosts[:filters][:products][key] : val
    #end
    #@hosts[:filters][:products] = products
  end

  DAY_SECONDS = 24 * 60 * 60
  DATE_INVALID = Date.new(4000,12,31).freeze
  DATE_NONE = Date.new(3000,12,31).freeze

  # フィルタの適用
  # 未使用。カスタムタグ検索未実装
  def apply_filters
    #履歴は対象のhtml_idを先に取得する
    is_filter_history = false
    filter_history_host_ids = []
    if @hosts[:filters][:journal][:period] != '::' then
      if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @hosts[:filters][:journal][:period] then
        orig_flg_rev = Regexp.last_match(1)
        orig_begin_date = Regexp.last_match(2)
        orig_end_date = Regexp.last_match(3)
        if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
          @hosts[:filters][:journal][:period] = '::'
        else
          flg_rev = orig_flg_rev == 'rev'
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          begin_date = begin_date.to_time if begin_date
          end_date = end_date.to_time + DAY_SECONDS - 1 if end_date
          is_filter_history = true

          filter_history_host_ids = Journal.joins(:issue)
                                           .joins('INNER JOIN sidfmd_issues ON sidfmd_issues.issue_id = issues.id')
                                           .joins('LEFT OUTER JOIN journal_details ON journal_details.journal_id = journals.id')
                                           .where('sidfmd_issues.html_id = ?', @id)
                                           .where('sidfmd_issues.sidfmd_host_id IN (?)', @hosts[:all_content].map{|h| h[:sidfmd_host_id]}) 
          if flg_rev then
            if begin_date and end_date then
              filter_history_host_ids = filter_history_host_ids.where.not(:journals => { :created_on => begin_date..end_date })
            elsif begin_date then
              filter_history_host_ids = filter_history_host_ids.where.not('journals.created_on >= ?', begin_date)
            elsif end_date then
              filter_history_host_ids = filter_history_host_ids.where.not('journals.created_on <= ?', end_date)
            else
              # last resort
              @hosts[:filters][:journal][:period] = '::'
              is_filter_history = false
            end
          else
            if begin_date and end_date then
              filter_history_host_ids = filter_history_host_ids.where(:journals => { :created_on => begin_date..end_date })
            elsif begin_date then
              filter_history_host_ids = filter_history_host_ids.where('journals.created_on >= ?', begin_date)
            elsif end_date then
              filter_history_host_ids = filter_history_host_ids.where('journals.created_on <= ?', end_date)
            else
              # last resort
              @hosts[:filters][:journal][:period] = '::'
              is_filter_history = false
            end
          end
          if @hosts[:filters][:journal][:filter_status_change] and @hosts[:filters][:journal][:filter_duedate_change] then
            f_host_ids = filter_history_host_ids.where(:journal_details => {:prop_key => 'status_id'}).pluck('sidfmd_issues.sidfmd_host_id') &
                         filter_history_host_ids.where(:journal_details => {:prop_key => 'due_date'}).pluck('sidfmd_issues.sidfmd_host_id')
            if f_host_ids.any? then
              filter_history_host_ids = filter_history_host_ids.where(%|
                sidfmd_issues.sidfmd_host_id IN (SELECT unnest(ARRAY[#{f_host_ids.join(', ')}]))
              |)
            else
              filter_history_host_ids = filter_history_host_ids.where(:journals => { :id => 0 })
            end
          elsif @hosts[:filters][:journal][:filter_status_change] then
            filter_history_host_ids = filter_history_host_ids.where(:journal_details => { :prop_key => 'status_id' })
          elsif
            @hosts[:filters][:journal][:filter_duedate_change] then
            filter_history_host_ids = filter_history_host_ids.where(:journal_details => { :prop_key => 'due_date' })
          end
          filter_history_host_ids = filter_history_host_ids.pluck('sidfmd_issues.sidfmd_host_id')
        end
      else
        @hosts[:filters][:journal][:period] = '::'
      end
    end
   
    ret = @hosts[:all_content].find_all {|item|
      # 要注意
      if @hosts[:filters][:warning] == 'warning' then
        next false unless item[:warning][@hosts[:idx]]
      elsif @hosts[:filters][:warning] == 'others' then
        next false if item[:warning][@hosts[:idx]]
      end
      # 対応状況
      if /\A(\d):(\d)\z/ =~ @hosts[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:status_id] == status1 or item[:status_id] == status2)
      elsif @hosts[:filters][:status] != '0' then
        next false unless item[:status_id] == @hosts[:filters][:status].to_i
      end
      # SRI
      if @hosts[:filters][:sris].any? then
        next false unless @hosts[:filters][:sris].include?(item[:sri].to_s)
      end
      # CVSS
      if @hosts[:filters][:cvsses].any? then
        next false unless @hosts[:filters][:cvsses].include?(item[:cvss][:label].to_s)
      end
      # 攻撃コード
      if @hosts[:filters][:exploitable] != '' then
        next false unless (@hosts[:filters][:exploitable] == 'true' && item[:exploitable])
      end
      #カテゴリ(複数)
      if @hosts[:filters][:categories][:ids].any? then
        pass = false
        host_all_cat_ids = item[:no_system_categories].map{|cc| cc[:id]}
        host_all_cat_ids << item[:system_category][:id] if item[:system_category]
        if @hosts[:filters][:categories][:is_and] then
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).empty? then
            pass = true
          end
        elsif @hosts[:filters][:categories][:is_except] then
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).size == @hosts[:filters][:categories][:ids].size then
            pass = true
          end
        else
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).size != @hosts[:filters][:categories][:ids].size then
            pass = true
          end
        end
        next false unless pass
      end
      # OS
      if @hosts[:filters][:os] != '' then
        next false unless (item[:os] and @hosts[:filters][:os] == item[:os][:osname])
      end
      # 履歴
      if is_filter_history then
        next unless filter_history_host_ids.include?(item[:sidfmd_host_id])
      end
=begin
      if @hosts[:filters][:journal] != '' then
        if @hosts[:filters][:journal] == 'exist' then
          next false unless item[:journals].size > 0 
        else
          temp_timespan = 1.days
          case @hosts[:filters][:journal]
          when '1day'   then temp_timespan = 1.days
          when '1week'  then temp_timespan = 1.weeks
          when '1month' then temp_timespan = 1.months
          end
          timespan = Time.current.ago(temp_timespan)
          next false unless item[:journals].find {|x|
            x[:created_on] >= timespan
          }
        end
      end
=end
      # 対策予定日設定済
      if @hosts[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @hosts[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless (item[:due_date].nil? or item[:due_date].size == 0)
          else
            next false if item[:due_date].nil?
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              log_exception(e)
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              log_exception(e)
              enddate = nil
            end
            due_date = nil
            begin
              due_date = Date.parse(item[:due_date])
            rescue Exception => e
              log_exception(e)
              due_date = nil
            end
            if begindate and enddate then
              next false unless (due_date >= begindate and due_date <= enddate)
            elsif begindate then
              next false unless due_date >= begindate
            elsif enddate then
              next false unless due_date <= enddate
            end
          end
        end
      end
      # キーワード検索
      if @hosts[:filters][:keyword].size > 0 then
        next false unless
          #Regexp.new('.*' + Regexp.escape(@hosts[:filters][:keyword]) + '.*',
          #           Regexp::IGNORECASE) =~ item[:title] or
          Regexp.new('.*' + Regexp.escape(@hosts[:filters][:keyword]) + '.*',
                     Regexp::IGNORECASE) =~ item[:hostname] or 
          Regexp.new('.*' + Regexp.escape(@hosts[:filters][:keyword]) + '.*',
                     Regexp::IGNORECASE) =~ item[:display_name]
      end
      # 日付指定検索
      if @hosts[:filters][:date] != '::' then
        if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @hosts[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @hosts[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false if begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if end_date >= item[:created_on][:value]
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless end_date >= item[:created_on][:value]
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          end
        else
          @hosts[:filters][:date] = '::'
          next false
        end
      end
      true
    }
    #ret.sort! {|x, y|
    #  x[:status_id] != y[:status_id] ?
    #  SidfmdIssue::SORT_PRIORITIES[x[:status_id]] <=> SidfmdIssue::SORT_PRIORITIES[y[:status_id]] :
    #  y[:created_on][:value] <=> x[:created_on][:value]
    #}
    case @hosts[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:warning][@hosts[:idx]] ? 500 : 100 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:warning][@hosts[:idx]] ? 500 : 100 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'status_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:status_id] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'status_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:status_id] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'sri_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:sri] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'sri_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:sri] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'cvss_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:cvss][:score] ? x[:cvss][:score] : 0.0 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'cvss_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:cvss][:score] ? x[:cvss][:score] : 0.0 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [
         x[:display_name].downcase,
         x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 25) : 40
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [
         x[:display_name].downcase,
         x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 25) : 40
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'elapseddays_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, (x[:elapsed_days] == '---' ? -1 : x[:elapsed_days]) ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'elapseddays_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, (x[:elapsed_days] == '---' ? -1 : x[:elapsed_days]) ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'hostname_asc' then
      ret.sort! {|x, y| x[:display_name].downcase <=> y[:display_name].downcase }
    when 'hostname_desc' then
      ret.sort! {|y, x| x[:display_name].downcase <=> y[:display_name].downcase }
    when 'categories_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'categories_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'os_asc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:display_name].downcase, osname.downcase, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'os_desc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:display_name].downcase, osname.downcase, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'journals_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:journals].size ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'journals_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:journals].size ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'duedate_asc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:due_date] ? x[:due_date] : '0000/00/00' ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'duedate_desc' then
      temp_ev = proc {|x|
        [ x[:display_name].downcase, x[:due_date] ? x[:due_date] : '0000/00/00' ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    end
    ret
  end

  #hostviewのフィルタの適用
  def apply_filters_to_host_view_content
    get_summary_prior_category unless @summary_prior_category

    @hosts[:all_content] = filter_vul_content(
      :content => @hosts[:all_content],
      :filters => @hosts[:filters],
      :vul_view => @vul_view,
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids,
      :index => @hosts[:idx],
      :corresp_required => @corresp_required
    )

    # sort
    @hosts[:all_content] = sort_vul_content(
      :content => @hosts[:all_content],
      :order => @hosts[:order],
      :vul_view => @vul_view,
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids,
      :prior_category => @summary_prior_category,
      :index => @hosts[:idx],
      :corresp_required => @corresp_required
    )

  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month
=begin
  # タイムラインのデータを返す
  def get_timeline_data
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @hosts[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    @hosts[:all_content].each do |item|
      ret[:timescale].each do |step|
        next if item[:created_on][:value] > step[:time]
        # issue
        issue_status_id = item[:status_id]
        # journals
        temp_j = item[:journals].collect {|x|
          if x[:created_on] <= step[:time] and x[:details] and
              x[:details].find {|x| x[:prop_key] == 'status_id' } then
            x
          else
            nil
          end
        }
        temp_j.compact!
        temp_j.sort! {|x, y|
          x[:created_on] == y[:created_on] ?
          x[:id] <=> y[:id] :
          x[:created_on] <=> y[:created_on] }
        journal_status_id = nil
        if temp_j.size > 0 then
          temp_j.last[:details].each do |i|
            if i[:prop_key] == 'status_id' then
              journal_status_id = i[:value].to_i
              break
            end
          end
        end
        count_index = TIMELINE_SPAN - 1 - step[:step]
        if journal_status_id then
          ret[:status_counts][journal_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(journal_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        elsif issue_status_id then
          ret[:status_counts][issue_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(issue_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        end
      end
    end

    ret
  end
=end

  def get_timeline_data_orig
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @hosts[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1 ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    @hosts[:all_content].each do |item|
      ret[:timescale].each do |step|
        next if item[:created_on][:value] > step[:time]
        # issue
        issue_status_id = item[:status_id]
        # journals
        temp_j = item[:journals].collect {|x|
          if x[:created_on] <= step[:time] and x[:details] and
              x[:details].find {|x| x[:prop_key] == 'status_id' } then
            x
          else
            nil
          end
        }
        temp_j.compact!
        temp_j.sort! {|x, y|
          x[:created_on] == y[:created_on] ?
          x[:id] <=> y[:id] :
          x[:created_on] <=> y[:created_on] }
        journal_status_id = nil
        if temp_j.size > 0 then
          temp_j.last[:details].each do |i|
            if i[:prop_key] == 'status_id' then
              journal_status_id = i[:value].to_i
              break
            end
          end
        end
        count_index = TIMELINE_SPAN - 1 - step[:step]
        if journal_status_id then
          ret[:status_counts][journal_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(journal_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          end
        elsif issue_status_id then
          ret[:status_counts][issue_status_id][count_index] += 1
          #if SidfmdIssue::STATUS_UNFINISHED.include?(issue_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          #end
        end
      end
    end

    ret
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :grp, :idx, :category_id, :filters, :checked, :all_checked, :is_update_one_row, :timeline_risk_period, :timeline_status_period)

    @hosts = { :url => request.url }
    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @hosts[:category_id] = (pages_params[:category_id] || 0).to_i
    @hosts[:category_name] = '全てのホスト'
    if @hosts[:category_id] > 0 and
        (@system_categories.where(:id => @hosts[:category_id]).size > 0 or
         @role_admin or
         (!@role_admin and @user_category_ids.include?(@hosts[:category_id]))) then
      cat = SidfmdCategory.find(@hosts[:category_id])
      @hosts[:category_name] = cat.name if cat
    else
      @hosts[:category_id] = 0
    end

    @hosts[:grp] = (pages_params[:grp] || 'count').to_sym
    @hosts[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @hosts[:idx]
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (@summary_table_lines || pages_params[:per_page] || 10).to_i
    @hosts[:order] = pages_params[:order] || 'hostname_asc'
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = (pages_params[:all_checked] || '').to_s
    @hosts[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    @hosts[:timeline_risk_period] = (pages_params[:timeline_risk_period] || "month").to_sym
    @hosts[:timeline_status_period] = (pages_params[:timeline_status_period] || "month").to_sym

    @corresp_required = Sidfm.corresp_required
    #create_all_content(@hosts[:category_id])
    create_host_view_content(@hosts[:category_id])
    if @vul_view then
      check_filters

      @target_scsids = SidfmdVvScsid.where(:sidfmd_host_id => @hosts[:all_content].select(:sidfmd_host_id))
                                    .where(:sidfmd_vul_view_id => @vul_view.id)
      warning_scsids = if @hosts[:idx] == :sri then
                         @target_scsids.where('sri >= ?', @corresp_required[:sri])
                       else
                         @target_scsids.where('cvss >= ? OR cvss IS NULL', @corresp_required[:cvss])
                       end
      @hosts[:corresp_required_host_size] = warning_scsids.count
      @hosts[:fixed_corresp_required_host_size] = warning_scsids.where(:status_id => SidfmdIssue::STATUS_FINISHED).count
      @hosts[:not_fixed_corresp_required_host_size] = @hosts[:corresp_required_host_size] - @hosts[:fixed_corresp_required_host_size]
      @hosts[:all_host_size] = @hosts[:all_content].count
      @is_update_one_row = (pages_params[:is_update_one_row] ? (pages_params[:is_update_one_row] == 'true' ? true : false) : false) 
      @hosts[:content_filtered] = apply_filters_to_host_view_content
      @hosts[:total_size] = @hosts[:content_filtered].count
    end
  end

  def filter_content
    @status = {
      :exploit => [ { :value => @hosts[:not_fixed_corresp_required_host_size] } ]
    }
    @hosts[:distribution_status] = {
      :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0,
    }
    @hosts[:distribution_risk] =
      case @hosts[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @hosts[:all_content].each do |i|
      case i[:status_id]
      when SidfmdIssue::STATUS_NOTCONFIRMED then
        @hosts[:distribution_status][:notconfirmed] += 1
      when SidfmdIssue::STATUS_ONGOING then
        @hosts[:distribution_status][:ongoing] += 1
      when SidfmdIssue::STATUS_FIXED then
        @hosts[:distribution_status][:fixed] += 1
      when SidfmdIssue::STATUS_NOTAFFECT then
        @hosts[:distribution_status][:notaffect] += 1
      end
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:status_id]) then
        case @hosts[:idx]
        when :sri then
          case i[:sri]
          when SidfmdIssueSri::VALUES[0] then
            @hosts[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @hosts[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @hosts[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @hosts[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @hosts[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:cvss][:label]
          when :notreg then
            @hosts[:distribution_risk][:notreg] += 1
          when :d4 then
            @hosts[:distribution_risk][:d4] += 1
          when :d3 then
            @hosts[:distribution_risk][:d3] += 1
          when :d2 then
            @hosts[:distribution_risk][:d2] += 1
          when :d1 then
            @hosts[:distribution_risk][:d1] += 1
          end
        end
      else
        @hosts[:distribution_risk][:fixed] += 1
      end
    end
    @hosts[:content_filtered] = apply_filters
    @hosts[:total_size] = @hosts[:content_filtered].size

    @hosts[:statuses] = SidfmdIssue.statuses

    @statuses = IssueStatus.all.order(:id)

    @hosts[:dist_cat_hosts] = {}
    [ :public_server, :private_server, :shared_server,
      :mobile, :client, :virtual_host_server, :public_network,
      :private_network ].each_with_index do |v, i|
      @hosts[:dist_cat_hosts][v] =
        @hosts[:all_content].sum {|x| (x[:system_category] ? (x[:system_category][:id] == i + 1 ? 1 : 0) : 0) }
    end
    @hosts[:dist_os_hosts] = {}
    @hosts[:all_content].each do |host|
      os =
        (host[:os] and  host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '（不明）'
      if @hosts[:dist_os_hosts].has_key?(os) then
        @hosts[:dist_os_hosts][os] += 1
      else
        @hosts[:dist_os_hosts][os] = 1
      end
    end

    #@timeline_data = (@hosts[:grp] == :distribution) ? get_timeline_data : []
  end
  
  def filter_host_view_content
    @status = {
      :exploit => [ { :value => @hosts[:not_fixed_corresp_required_host_size] } ]
    }
    @hosts[:distribution_status] = {
      :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0,
    }
    @hosts[:distribution_risk] =
      case @hosts[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    # status
    @hosts[:distribution_status][:notconfirmed] = @target_scsids.where(:status_id => SidfmdIssue::STATUS_NOTCONFIRMED ).count
    @hosts[:distribution_status][:ongoing] = @target_scsids.where(:status_id => SidfmdIssue::STATUS_ONGOING ).count
    @hosts[:distribution_status][:fixed] = @target_scsids.where(:status_id => SidfmdIssue::STATUS_FIXED ).count
    @hosts[:distribution_status][:notaffect] = @target_scsids.where(:status_id => SidfmdIssue::STATUS_NOTAFFECT ).count
    # risk
    status_unfinished_hosts = @target_scsids.where(:status_id => SidfmdIssue::STATUS_UNFINISHED)
    if @hosts[:idx] == :sri then
      @hosts[:distribution_risk][:critical] = status_unfinished_hosts.where(:sri => SidfmdIssueSri::VALUES[0]).count
      @hosts[:distribution_risk][:high] = status_unfinished_hosts.where(:sri => SidfmdIssueSri::VALUES[1]).count
      @hosts[:distribution_risk][:medium] = status_unfinished_hosts.where(:sri => SidfmdIssueSri::VALUES[2]).count
      @hosts[:distribution_risk][:low] = status_unfinished_hosts.where(:sri => SidfmdIssueSri::VALUES[3]).count
      @hosts[:distribution_risk][:none] = status_unfinished_hosts.where(:sri => SidfmdIssueSri::VALUES[4]).count
    else
      @hosts[:distribution_risk][:notreg] = status_unfinished_hosts.where(:cvss => nil).count
      @hosts[:distribution_risk][:d4] = status_unfinished_hosts.where('cvss >= ?', SidfmdIssueCve::LABEL[:d4][:bottom]).count
      @hosts[:distribution_risk][:d3] = status_unfinished_hosts.where('cvss < ?', SidfmdIssueCve::LABEL[:d3][:top])
                                                               .where('cvss >= ?', SidfmdIssueCve::LABEL[:d3][:bottom]).count
      @hosts[:distribution_risk][:d2] = status_unfinished_hosts.where('cvss < ?', SidfmdIssueCve::LABEL[:d2][:top])
                                                               .where('cvss >= ?', SidfmdIssueCve::LABEL[:d2][:bottom]).count
      @hosts[:distribution_risk][:d1] = status_unfinished_hosts.where('cvss < ?', SidfmdIssueCve::LABEL[:d1][:top]).count
    end
    @hosts[:distribution_risk][:fixed] = @target_scsids.where.not(:status_id => SidfmdIssue::STATUS_UNFINISHED).count
   
    @hosts[:statuses] = SidfmdIssue.statuses

    @statuses = IssueStatus.all.order(:id)

    @hosts[:dist_cat_hosts] = {}
    [ :public_server, :private_server, :shared_server,
      :mobile, :client, :virtual_host_server, :public_network,
      :private_network ].each_with_index do |v, i|
      @hosts[:dist_cat_hosts][v] = @hosts[:all_content].where(:system_category_id => i + 1).count
    end
    @hosts[:dist_os_hosts] = SidfmdHostView.where(:id => @hosts[:all_content].select(:id)).group(:osname).count
    if @hosts[:dist_os_hosts][nil] then
      undef_osname_count = @hosts[:dist_os_hosts][nil]
      @hosts[:dist_os_hosts].delete(nil)
      @hosts[:dist_os_hosts]['（不明）'] = undef_osname_count
    end
  end

  def create_hosts_journals_count(content)
    @hosts_journals_count = {}
    content.each do |host|
      sidfmd_issue = SidfmdIssue.where(:sidfmd_host_id => host[:sidfmd_host_id])
                                .where(:html_id => @html[:html_id])
                                .first
      if sidfmd_issue then
        @hosts_journals_count[host[:sidfmd_host_id]] = Journal.where(:journalized_id => sidfmd_issue.issue_id).count
      else
        @hosts_journals_count[host[:sidfmd_host_id]] = 0
      end
    end
  end
  
  def get_host_ctags_data(host_view_ids)
    ctags = SidfmdCtag.joins(:sidfmd_hv_ctags)
      .where(:sidfmd_hv_ctags => {
        :sidfmd_host_view_id => host_view_ids,
        :html_id => @id 
      })
      .order(%| sidfmd_ctags.priority DESC |)
      .order(%| LOWER(sidfmd_ctags.keyword COLLATE "C") |)
    unless @role_admin then
      ctags = ctags.where("sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (?)",  @user_category_ids)
    end
    ctags
  end

  #ホスト毎のカスタムタグ
  def create_host_ctags(host_view_ids)
    @host_ctags = {}
    ctags = get_host_ctags_data(host_view_ids)
    ctags.select(
      "sidfmd_ctags.keyword AS tag_name",
      "sidfmd_ctags.color AS text_color",
      "sidfmd_ctags.bgcolor AS bg_color",
      "sidfmd_hv_ctags.sidfmd_host_view_id AS view_id"
    ).each do |ctag|
      @host_ctags[ctag.view_id] = [] unless @host_ctags[ctag.view_id]
      @host_ctags[ctag.view_id] << {
        :keyword => ctag.tag_name,
        :text_color => ctag.text_color,
        :bg_color => ctag.bg_color
      }
    end
  end

  #JSON/TSVファイル出力のカスタムタグ
  def create_host_ctags_for_file(host_view_ids)
    @host_ctags_for_file = {}
    ctags = get_host_ctags_data(host_view_ids)
    ctags.select(
      "sidfmd_ctags.keyword",
      "sidfmd_ctags.sidfmd_category_id",
      "sidfmd_ctags.color",
      "sidfmd_ctags.bgcolor",
      "sidfmd_ctags.priority",
      "sidfmd_hv_ctags.sidfmd_host_view_id AS view_id"
    ).each do |ctag|
      @host_ctags_for_file[ctag.view_id] = [] unless @host_ctags_for_file[ctag.view_id]
      @host_ctags_for_file[ctag.view_id] << {
        :keyword => ctag.keyword,
        :category_id => ctag.sidfmd_category_id,
        :color => ctag.color,
        :background_color => ctag.bgcolor,
        :priority => ctag.priority
      }
    end
  end
 
  def create_summary_host_html(host_view, **opts)
    sidfmd_issue = SidfmdIssue.find_by(:html_id => @vul_view.html_id, :sidfmd_host_id => host_view.sidfmd_host_id)
    is_warning =
      if SidfmdIssue::STATUS_UNFINISHED.include?(opts[:scsid].status_id) then
        if opts[:idx] == :sri then
          (opts[:scsid].sri >= opts[:corresp_required][:sri])
        else
          (opts[:scsid].cvss.nil? or opts[:scsid].cvss >= opts[:corresp_required][:cvss])
        end
      else
        false
      end
    cvss_label = SidfmdIssueCve.labelling(opts[:scsid].cvss).to_s
    data = {
      :chkbox => view_context.create_summary_host_chkbox(host_view.sidfmd_host_id, opts[:checked]),
      :warning => view_context.create_summary_taiou(is_warning),
      :status => view_context.create_summary_status([opts[:scsid].status_id], opts[:scsid].status_id, false),
      :sri => view_context.create_summary_vul_sri(opts[:scsid].sri, @vul_view.html_id),
      :cvss => view_context.create_summary_vul_detail_cvss(cvss_label, opts[:scsid].cvss, @vul_view.html_id),
      :area => view_context.create_summary_vul_area(@vul_view.softek_area_id, @vul_view.html_id),
      :exploitable => view_context.create_summary_detail_exploit(@vul_view, opts[:scsid].status_id, cvss_label),
      :elapseddays => view_context.create_summary_vul_detail_elapsed_days(opts[:scsid].created_on, opts[:scsid].status_id),
      :hostname => view_context.create_summary_vul_detail_hostname(host_view),
      :os => view_context.create_summary_host_osname(host_view, '不明'),
      :categories => view_context.create_summary_host_categories(
        host_view.sidfmd_host_id,
        host_view.system_category_id,
        host_view.system_category_name,
        (
          @role_admin ?
          host_view.sidfmd_hv_categories :
          host_view.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
        ).order('LOWER(name COLLATE "C")'),
        opts[:summary_prior_category],
        { :html_id => @vul_view.html_id }
      ),
      :detail => view_context.create_summary_vul_detail_affected_items(@vul_view, sidfmd_issue),
      :journals => view_context.create_summary_vul_detail_journal(
        @vul_view,
        host_view,
        (sidfmd_issue ? Journal.where(:journalized_id => sidfmd_issue.issue_id).count : 0),
        opts[:user_id]
      ),
      :duedate => view_context.create_summary_duedate(opts[:scsid].due_date, opts[:scsid].status_id),
      :ctags => view_context.create_summary_custom_tag(
        (@host_ctags[host_view.id] ? @host_ctags[host_view.id].first : nil),
        @vul_view.html_id,
        host_view.display_name,
        @sidfm_roles,
        { :sidfmd_host_id => host_view.sidfmd_host_id }
      ),
      :action => view_context.create_summary_detail_action(
        host_view,
        opts[:scsid].status_id,
        "#modalEditStatus",
        "#modalEditSchedule",
        'host'
      )
    }
  end
  
  public

  def index
    f_params = params.fetch(:form, {}).permit(:history)
    get_summary_prior_category
    parse_params
    #filter_content
    if !@vul_view then
      redirect_to :controller => :sidfm_vulnerabilities, :action => :index_r, :notice => '指定された脆弱性は存在しません。' and return
    end
    filter_host_view_content
    if session['hostid_with_htmlid'] then
      if session['hostid_with_htmlid']['host'] then
        @hosts[:content_filtered].each do |host|
          h_id = host[:sidfmd_host_id].to_s
          if session['hostid_with_htmlid']['host'][h_id] then
            session['hostid_with_htmlid']['host'][h_id][1] = true
          end
        end
      end
      if session['hostid_with_htmlid']['vul'] && !@is_update_one_row then
        session['hostid_with_htmlid']['vul'].each_key do |htmlid|
          session['hostid_with_htmlid']['vul'][htmlid][1] = false
        end
      end
    end

    @history_filter_host = nil
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    if f_params[:history] then
      history = JSON.parse(f_params[:history])
      host_id = history[0].to_i
      html_id = history[1].to_i
      @history_filter_host = @hosts[:content_filtered].where(:sidfmd_host_id => host_id).select(
        :sidfmd_host_id,
        :hostname
      ).first
      #count = 0
      #@hosts[:content_filtered].each do |host|
      #  if host[:sidfmd_host_id] == host_id then
      #    break
      #  else
      #    count += 1;
      #  end
      #end
      #@hosts[:offset] = count
      session[:confirming_journal] = {:host_id => host_id, :html_id => html_id}
    end
    #@hosts[:content] = @hosts[:content_filtered].slice(@hosts[:offset], @hosts[:per_page])
    @hosts[:content] = @hosts[:content_filtered].offset(@hosts[:offset]).limit(@hosts[:per_page])
    @hosts[:content] = [] unless @hosts[:content]

    #host_ids = []
    #journalの数をキャッシュせずに反映させる
    create_hosts_journals_count(@hosts[:content]) 
    #@hosts[:content].each do |host|
    #  host_ids << host[:sidfmd_host_id]
    #  sidfmd_issue = SidfmdIssue.where(:sidfmd_host_id => host[:sidfmd_host_id])
    #                            .where(:html_id => @html[:html_id])
    #  if sidfmd_issue.size > 0 then
    #    @hosts_journals_count[host[:sidfmd_host_id]] = Journal.where(:journalized_id => sidfmd_issue.first.issue_id).count
    #  else
    #    @hosts_journals_count[host[:sidfmd_host_id]] = 0
    #  end             
    #end

    @display_not_system_categories = SidfmdCategory.where(:id => @not_system_categories.map{|c| c.id}).order('name COLLATE "C"')
    pvt_filters = SidfmdPvtFilter.get_filters(SidfmdGrpFilter::KEYWORD_VUL, User.current.id)
    grp_filters = SidfmdGrpFilter.get_filters(SidfmdGrpFilter::KEYWORD_VUL, (@role_admin ? nil : @user_category_ids))
    @stored_filters = pvt_filters + grp_filters 
    #ローカル脆弱性の影響するローカルプロダクト
=begin
    @issue_lproduct_names = {} 
    lproducts = SidfmdLocalVulLapp.joins(:sidfmd_local_vul).joins(
      'LEFT OUTER JOIN sidfmd_issues ON sidfmd_local_vuls.html_id = sidfmd_issues.html_id'
    )
    .joins('LEFT OUTER JOIN sidfmd_host_lapps ON sidfmd_issues.sidfmd_host_id = sidfmd_host_lapps.sidfmd_host_id')
    .where("sidfmd_issues.html_id = #{@html[:html_id]}")
    .where('sidfmd_host_lapps.sidfmd_host_id IN (?)', host_ids)
    .where('sidfmd_host_lapps.sidfmd_lproduct_id IS NOT NULL')
    .select(
      'sidfmd_issues.html_id AS html_id',
      'sidfmd_host_lapps.sidfmd_lproduct_id AS host_lapp_lpid',
      'sidfmd_host_lapps.sidfmd_lversion_id AS host_lapp_lvid',
      'sidfmd_host_lapps.sidfmd_host_id AS host_id'
    )
   
    if lproducts.any? then
      local_vul_lpids_of_all = {}
      local_vul_lvids = {}
      local_vul_lpids = {}
      lpids = []
      SidfmdLocalVul.joins(:sidfmd_local_vul_lapps)
                    .select(
                      'sidfmd_local_vul_lapps.sidfmd_lproduct_id AS lpid',
                      'sidfmd_local_vul_lapps.sidfmd_lversion_id AS lvid',
                      'sidfmd_local_vuls.html_id AS html_id'
                    )
                    .each do |lapp|
                      if lapp.lvid == 0 then #バージョンが「全て」の場合
                        local_vul_lpids_of_all[lapp.html_id] = [] unless local_vul_lpids_of_all[lapp.html_id]
                        local_vul_lpids_of_all[lapp.html_id] << lapp.lpid
                      else #バージョンが「全て」でない場合
                        local_vul_lvids[lapp.html_id] = [] unless local_vul_lvids[lapp.html_id]
                        local_vul_lvids[lapp.html_id] << lapp.lvid
                      end
                      local_vul_lpids[lapp.html_id] = [] unless local_vul_lpids[lapp.html_id]
                      local_vul_lpids[lapp.html_id] << lapp.lpid
                      lpids << lapp.lpid
                    end
      lprod_names = {}
      lver_names = {}
      SidfmdLproduct.where(:id => lpids).each{|lp| lprod_names[lp.id] = lp.name}
      SidfmdLversion.where(:sidfmd_lproduct_id => lpids).each{|lv| lver_names[lv.id] = lv.fullname}

      lproducts.each do |lapp|
        @issue_lproduct_names[lapp.host_id] = [] unless @issue_lproduct_names[lapp.host_id]
        if !lapp.host_lapp_lvid || lapp.host_lapp_lvid == 0 then #バージョンが「全て」の場合
          lprod_ids = local_vul_lpids[lapp.html_id]
          if lprod_ids && lprod_ids.include?(lapp.host_lapp_lpid) then
            @issue_lproduct_names[lapp.host_id] << lprod_names[lapp.host_lapp_lpid]
          end
        else #バージョンが「全て」でない場合
          lver_ids = local_vul_lvids[lapp.html_id]
          if lver_ids && lver_ids.include?(lapp.host_lapp_lvid) then
            @issue_lproduct_names[lapp.host_id] << lver_names[lapp.host_lapp_lvid]
          else
            lprod_ids = local_vul_lpids_of_all[lapp.html_id]
            if lprod_ids && lprod_ids.include?(lapp.host_lapp_lpid) then
              @issue_lproduct_names[lapp.host_id] << lver_names[lapp.host_lapp_lvid]
            end
          end
        end
      end
    end
=end

    @deleted_comment = HISTORY_COMMENT_DELETED_NOTE
    #ブラウザバックかどうか判断するための処理
    #if !session['chkbok_ids'] then
    #  session['chkbok_ids'] = {}
    #end
    #if !session['chkbok_ids']['vul'] then
    #  session['chkbok_ids']['vul'] = {}
    #end
    #session['chkbok_ids']['vul']['is_browser_back'] = false
  end

  def update_status
=begin    f_params = params.fetch(:status, []).permit(:hostids, :prev_status, :new_status, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !hostids.respond_to?(:'[]') or !hostids.respond_to?(:size) or hostids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if new_status == 0 then
      dump[:notice] = '設定後ステータスが設定されなかったため、操作は行われておりません。'
      redirect_to dump and return
    end
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
          .where(:html_id => @id)
        #if prev_status == 0 then
        #  issues = issues
        #    .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        # else
        #  issues = issues
        #    .where(:issues => { :status_id => prev_status })
        #end
        #issues.each do |issue|
        #  issue.with_lock do
        #    issue.change_status_with_journal(User.current.id, new_status, comment)
        #  end
        #end
        if prev_status > 0 then
          issues = issues.where(:issues => { :status_id => prev_status })
        end
        SidfmdIssue.change_statuses_with_journal(issues.map{|i|i.id}, nil, new_status, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end
    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids,:shift_num, :ids, :prev_status, :new_status, :comment, :action_name)
    err = false
    err_info = {}
    hostids = nil
    shift_num = f_params[:shift_num]
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      hostids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      err = true
    end
    if !hostids.respond_to?(:'[]') or !hostids.respond_to?(:size) or hostids.size == 0 then
      err = ture
    end
    if !@role_admin then
      if (hostids - @user_host_ids).size > 0 then
        err = true
      end
      hostids = hostids & @user_host_ids
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      err = true
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      err = true
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if new_status == 0 then
      err = true
    end
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    #target_html_ids = []
    #modify_host_ids = []
    update_data = []
    begin
      SidfmdIssue.transaction do
        old_status_id = nil
        if prev_status > 0 then
          old_status_id = prev_status
        end
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .joins(:issue)
            .where(:sidfmd_host_id => hostids)
            .where(:html_id => @id)
          if prev_status > 0 then
            issues = issues.where(:issues => { :status_id => prev_status })
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        SidfmdIssue.where(:issue_id => p_issue_ids).each do |i|
          update_data << {
            :sidfmd_issue_id => i.id,
            :html_id => i.html_id,
            :sidfmd_host_id => i.sidfmd_host_id,
            :status_id => new_status
          }
        end
        #target_html_ids = SidfmdIssue.where(:issue_id => p_issue_ids).distinct.pluck(:html_id)
        #modify_host_ids = SidfmdVulView.relate_hostids(target_html_ids)
        SidfmdIssue.change_statuses_with_journal(p_issue_ids, old_status_id, new_status, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      logger.info(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.update_from_rule_action(update_data)
      SidfmdHostView.update_from_rule_action(update_data)
      #modify_host_ids |= SidfmdVulView.relate_hostids(target_html_ids)
      #SidfmdVulView.sync_data(target_html_ids)
      #SidfmdHostView.sync_data(modify_host_ids)
    end

    action = f_params[:action_name]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
    end
    render :layout => false, :json => return_value
  end

  def update_due_dates
=begin    f_params = params.fetch(:schedule, {}).permit(:hostids, :force, :due_date, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if hostids.size == 0 or !hostids.respond_to?(:'[]') then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
          .where(:html_id => @id)
        if !force then
          issues = issues
            .where('issues.due_date IS NULL')
        end
        #issues.each do |issue|
        #  issue.with_lock do
        #    issue.change_due_date_with_journal(User.current.id, due_date, comment)
        #  end
        #end
        issue_ids = issues.map{|i| i.issue.id}
        SidfmdIssue.change_due_dates_with_journal(issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids, :shift_num, :ids, :force, :due_date, :comment, :action_name)
    err = false
    err_info = {}
    shift_num = f_params[:shift_num]
    hostids = nil
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      hostids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      logger.info(e)
      err = true
    end
    if hostids.size == 0 or !hostids.respond_to?(:'[]') then
      err = true
    end
    if !@role_admin then
      if (hostids - @user_host_ids).size > 0 then
        err = true
      end
      hostids = hostids & @user_host_ids
    end
    force = f_params[:force]  == 'true' ? true : false
    due_date = f_params[:due_date]
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        logger.info(e)
        err = true
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          logger.info(e)
          err = true
        end
      end
      if temp then
        due_date = temp
      else
        err = true
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    #target_html_ids = []
    #modify_host_ids = []
    update_data = []
    begin
      SidfmdIssue.transaction do
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .joins(:issue)
            .where(:sidfmd_host_id => hostids)
            .where(:html_id => @id)
          if !force then
            issues = issues.where('issues.due_date IS NULL')
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        SidfmdIssue.where(:issue_id => p_issue_ids).each do |i|
          update_data << {
            :sidfmd_issue_id => i.id,
            :html_id => i.html_id,
            :sidfmd_host_id => i.sidfmd_host_id,
            :due_date => due_date
          }
        end
        #target_html_ids = SidfmdIssue.where(:issue_id => p_issue_ids).distinct.pluck(:html_id)
        #modify_host_ids = SidfmdVulView.relate_hostids(target_html_ids)
        SidfmdIssue.change_due_dates_with_journal(p_issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.update_from_rule_action(update_data)
      SidfmdHostView.update_from_rule_action(update_data)
      #modify_host_ids |= SidfmdVulView.relate_hostids(target_html_ids)
      #SidfmdVulView.sync_data(target_html_ids)
      #SidfmdHostView.sync_data(modify_host_ids)
    end

    action = f_params[:action_name]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
    end
    render :layout => false, :json => return_value
  end

  def update_ctags
    if !@sidfm_roles.include?(:use) then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    f_params = params.permit(:html_id, :ctag_ids, :sidfmd_host_id)
    if !f_params[:sidfmd_host_id] or !f_params[:html_id] or !f_params[:ctag_ids] or f_params[:html_id].to_i != @id then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    begin
      ctag_ids = JSON.parse(f_params[:ctag_ids])
    rescue JSON::ParserError => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    begin
      issues = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:html_id => f_params[:html_id].to_i)
        .where(:sidfmd_host_id => f_params[:sidfmd_host_id].to_i)
      SidfmdIssueCtag.udpate_issue_ctags(
        ctag_ids: ctag_ids,
        sidfmd_issue_rows: issues,
        role_admin: @role_admin,
        user_category_ids: @user_category_ids,
        user_host_ids: @user_host_ids
      )
    rescue Exception => e
      logger.info(e)
      render :layout => false, :json => { :err_message => 'カスタムタグの更新に失敗しました。' } and return
    end

    render :layout => false, :json => { :err_message => '' }

  end

  def sendmail
    f_params = params.fetch(:sm, {}).permit(:title, :body, :htmlids)
    logger.info(f_params)
    redirect_to :action => :index
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  MAX_CELL_LENGTH = 30000
  MAX_LINE_NUM = 254 # =最大改行数は253

  def read_download_data
    GC.start
    f_params = params.permit(:target, :executed_size, :process, :category_id, :idx, :order, :filters, :ids)
    if !f_params[:ids] or !f_params[:target] or !f_params[:order] or !f_params[:filters] or
       !f_params[:executed_size] or !f_params[:process] or !f_params[:idx] or !f_params[:category_id]
    then
      logger.info('params error')
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    format = f_params[:target]
    if ![ 'json', 'tsv', 'tsv-package-info', 'tsv-affect-products' ].include?(format) then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    executed_size = f_params[:executed_size].to_i
    get_summary_prior_category

    @hosts = { :url => request.url }
    @hosts[:category_id] = (f_params[:category_id] || 0).to_i
    @hosts[:category_name] = '全てのホスト'
    if @hosts[:category_id] > 0 then
      cat = SidfmdCategory.find(@hosts[:category_id])
      @hosts[:category_name] = cat.name if cat
    end
    @hosts[:idx] = (f_params[:idx] || 'sri').to_sym
    @hosts[:order] = f_params[:order] || 'hostname_asc'
    @hosts[:filters] = JSON.parse((f_params[:filters] || '{}'), symbolize_names: true)
    @hosts[:hostids] = []
    begin
      @hosts[:ids] = JSON.parse(f_params[:ids] || '[]') #processがinitの時(最初の時)はsidfmd_host_idが入り、2回目以降はSidfmdHostVeiwのidが入ることに注意
    rescue JSON::ParserError => e
      logger.error(e)
    end
    @corresp_required = Sidfm.corresp_required
    create_host_view_content(@hosts[:category_id])
    @target_scsids = SidfmdVvScsid.where(:sidfmd_host_id => @hosts[:all_content].select(:sidfmd_host_id))
                                  .where(:sidfmd_vul_view_id => @vul_view.id)
    check_filters
    @hosts[:content_filtered] = apply_filters_to_host_view_content

    case format
    when 'json' then
      if f_params[:process] == 'init' then
        @hosts[:content_filtered] = @hosts[:content_filtered].where(:sidfmd_host_id => @hosts[:ids]) if @hosts[:ids].any?
        temp_order = @hosts[:order].to_s.split(/_/)
        init_json_summary_table_progress_download(
          :ids => @hosts[:content_filtered].pluck(:id),
          :order_col_name => temp_order[0],
          :order_type => temp_order[1],
          :filters => @hosts[:filters],
          :max_download_size => @hosts[:content_filtered].count,
          :category => { :id => @hosts[:category_id], :name => @hosts[:category_name] },
          :idx => @hosts[:idx]
        )
      else
        create_host_ctags_for_file(@hosts[:ids])
        create_json_data = Proc.new{|id|
          host = @hosts[:content_filtered].find_by(:id => id)
          next unless host
          issue = SidfmdIssue.where(:sidfmd_host_id => host.sidfmd_host_id, :html_id => @vul_view.html_id).first
          host_scsid = @target_scsids.find_by(:sidfmd_host_id => host.sidfmd_host_id)
          next if !issue or !host_scsid
          item = {}
          ctags = (@host_ctags_for_file[id.to_i] || [])
          item[:host] = {
            :id => host[:sidfmd_host_id],
            :hostname => host[:hostname],
            :ext_name => host[:ext_name],
            :display_name => host[:display_name]
          }
          item[:warning] = if SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid.status_id) then
                             if @hosts[:idx] == :sri then
                               if host_scsid.sri >= @corresp_required[:sri] then
                                 true
                               else
                                 false
                               end
                             else
                               if host_scsid.cvss.nil? || host_scsid.cvss >= @corresp_required[:cvss] then
                                 true
                               else
                                 false
                               end
                             end
                           else
                             false
                           end
          item[:status] = SidfmdIssue.status_id_to_s(host_scsid[:status_id])
          item[:sri] = SidfmdIssueSri::RISKS[host_scsid[:sri]]
          cvss_label = SidfmdIssueCve.labelling(host_scsid[:cvss]).to_s
          item[:cvss] = (cvss_label == 'notreg' ? '' : sprintf('%.1f', (host_scsid[:cvss] ? host_scsid[:cvss] : 0).to_f))
          item[:area] =  SidfmdArea.where(:area_id => @vul_view[:softek_area_id]).first.name
          item[:exploitable] = (cvss_label != 'notreg' and SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid[:status_id])) ? @vul_view[:exploitable] : nil
          ns_cats = (
            @role_admin ?
            host.sidfmd_hv_categories :
            host.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
          )
          .order('LOWER(name COLLATE "C")')
          .pluck(:name)
          cats = (
            @summary_prior_category == 'custom' ?
            ns_cats + [ host[:system_category_name] ] :
            [ host[:system_category_name] ] + ns_cats
          )
          item[:categories] = cats
          item[:os] = ((host[:osname] and host[:osname].size > 0) ? host[:osname] : '')
          item[:journals] =  []
          Journal.where(:journalized_id => issue.issue_id).order(:created_on => :desc).each do |journal|
            j_item = {}
            j_item[:id] = journal[:id]
            j_item[:created_on]  = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : '---'
            user = User.find(journal[:user_id])
            j_item[:user] = { :id => journal[:user_id], :name => user.anonymous? ? '削除済みユーザ' : view_context.display_user_name(user) }
            j_item[:notes] = Journal.find(journal[:id]).notes
            j_item[:details] = []
            JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
              jd_item = {}
              jd_item[:id] = detail[:id]
              case detail[:prop_key]
              when 'status_id' then
                jd_item[:action] = 'ステータス変更'
                jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
                jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
              when 'due_date' then
                jd_item[:action] = '対策予定日の変更'
                jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
                jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
              end
              j_item[:details] << jd_item
            end
            item[:journals] << j_item
          end
          item[:due_date] = host_scsid[:due_date] ? host_scsid[:due_date].strftime('%Y/%m/%d') : nil
          apps = []
          pkgs = []
          happs = []
          hlapps = []
          sidfmd_host = SidfmdHost.find_by(:id => host[:sidfmd_host_id])
          if sidfmd_host then
            queue = sidfmd_host.get_last_queue
            if queue then
              pkgs = queue.sidfmd_queue_pkgs.order(:pkg).collect {|x| x.pkg }
              apps = queue.sidfmd_queue_apps.order(:name,:version).collect {|x|
                { :name => x.name, :version => x.version, :path => x.path }
              }
            end
            happs = sidfmd_host.sidfmd_host_apps.order(:product_id,:version_id).collect {|x| x.get_name }.compact
            hlapps = sidfmd_host.sidfmd_host_lapps.order(:sidfmd_lproduct_id, :sidfmd_lversion_id).collect {|x| x.get_name }.compact
          end
          item[:packages] = pkgs
          item[:products] = apps
          item[:manual_products] = happs
          item[:local_products] = hlapps

          item[:affect_packages] = []
          item[:affect_products] = []

          item[:issue] = { :id => issue.id }
          item[:affect_packages] = issue.affect_packages('')
          item[:affect_products] = issue.affect_products + issue.affect_lproducts
          item[:ctags] = ctags
          item[:issue][:title] = issue.title
          item[:issue][:html_id] = issue.html_id > 0 ? "#{issue.html_id}" : "KB#{-issue.html_id}"
          item
        }
        json_tsv_summary_table_progress_download(@hosts[:ids], create_json_data, executed_size, 50)
      end
    when 'tsv' then
      convert_chars = ->(text){
        text.gsub('"', '”').gsub("'", "‘").gsub(',', '，').gsub("\n", ' ')
      }
      if f_params[:process] == 'init' then
        @hosts[:content_filtered] = @hosts[:content_filtered].where(:sidfmd_host_id => @hosts[:ids]) if @hosts[:ids].any?
        header = [ 'ホストID', '要対応', '対応状況', 'SRI', 'CVSS', 'CVE ID', 'CVSS V2', 'CVSS V3', '影響範囲', '攻撃コード', 'ホスト名', '表示名', 'カテゴリ', 'OS名', '履歴数', '履歴(日時,ユーザ名,メールアドレス,操作,操作内容,コメント)', '対策予定日', 'カスタムタグ' ]
        init_tsv_summary_table_progress_download(@hosts[:content_filtered].pluck(:id), header, @hosts[:content_filtered].count)
      else
        create_host_ctags_for_file(@hosts[:ids])
        create_tsv_data = Proc.new{|id|
          host = @hosts[:content_filtered].find_by(:id => id)
          next unless host
          issue = SidfmdIssue.where(:sidfmd_host_id => host.sidfmd_host_id, :html_id => @vul_view.html_id).first
          host_scsid = @target_scsids.find_by(:sidfmd_host_id => host.sidfmd_host_id)
          next if !issue or !host_scsid
          item = []
          ctags = (@host_ctags_for_file[id.to_i] || [])
          item << host[:sidfmd_host_id].to_s
          item << (if SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid.status_id) then
                    if @hosts[:idx] == :sri then
                      if host_scsid.sri >= @corresp_required[:sri] then
                        true
                      else
                        false
                      end
                    else
                      if host_scsid.cvss.nil? || host_scsid.cvss >= @corresp_required[:cvss] then
                        true
                      else
                        false
                      end
                    end
                  else
                    false
                  end).to_s
          item << SidfmdIssue.status_id_to_s(host_scsid[:status_id])
          item << SidfmdIssueSri::RISKS[host_scsid[:sri]]
          cvss_label = SidfmdIssueCve.labelling(host_scsid[:cvss]).to_s
          item << (cvss_label == 'notreg' ? '' : sprintf('%.1f', (host_scsid[:cvss] ? host_scsid[:cvss] : 0).to_f))
          cvss_2_cves = []
          cvss_3_cves = []
          cve_ids = []
          @vul_view.get_summary_cves.each do |cve|
            cve_ids << cve.cve_id
            cvss_2_cves << "#{cve.cve_id}|#{cve.c2_base_score}|#{cve.c2_vector_index}" if cve.c2_av
            cvss_3_cves << "#{cve.cve_id}|#{cve.c3_base_score}|#{cve.c3_vector_index}" if cve.c3_av
          end
          item << cve_ids.join(',')
          item << cvss_2_cves.join(',')
          item << cvss_3_cves.join(',')
          item << SidfmdArea.where(:area_id => @vul_view[:softek_area_id]).first.name
          item << ((cvss_label != 'notreg' and SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid[:status_id])) ? @vul_view[:exploitable] : nil).to_s
          item << host[:hostname]
          item << host[:display_name]
          ns_cats = (
            @role_admin ?
            host.sidfmd_hv_categories :
            host.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
          )
          .order('LOWER(name COLLATE "C")')
          .pluck(:name)
          cats = (
            @summary_prior_category == 'custom' ?
            ns_cats + [ host[:system_category_name] ] :
            [ host[:system_category_name] ] + ns_cats
          ).map{|cat_name| escape_category_tsv(cat_name) }
          item << cats.join(',')
          item << ((host[:osname] and host[:osname].size > 0) ? host[:osname] : '')
          host_journals = Journal.where(:journalized_id => issue.issue_id).order(:created_on)
          host_journals_size = host_journals.count 
          item << host_journals_size.to_s
          count_journal_lines = 0
          journal_text_size = 0
          journal_text_items = []
          host_journals.each_with_index do |journal, idx|
            j_text = ''
            if journal[:created_on] then
              j_text += journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') + ','
            end
            user = User.find(journal[:user_id])
            j_text += user.anonymous? ?  '削除済みユーザ,' : convert_chars.(view_context.display_user_name(user)) + ','
            j_text += user.mail ? "#{user.mail}," : ','
            detail_flag = false
            JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
              case detail[:prop_key]
              when 'status_id' then
                detail_flag = true
                j_text += 'ステータス変更' + ','
                j_text += "[#{SidfmdIssue.status_id_to_s(detail[:old_value].to_i)}]から[#{SidfmdIssue.status_id_to_s(detail[:value].to_i)}]へ,"
              when 'due_date' then
                detail_flag = true
                j_text += '対策予定日の変更' + ','
                j_text += "[#{SidfmdIssue.due_date_to_s(detail[:old_value])}]から[#{SidfmdIssue.due_date_to_s(detail[:value])}]へ,"
              end
            end
            if !detail_flag then
              j_text += ',,'
            end
            j_text += convert_chars.(Journal.find(journal[:id]).notes)
            j_text_size = j_text.size
            if host_journals_size != 1 then
              if idx != 0 then
                j_text_size += 1 #改行文字をカウント
              end
            end
            total_text_size = journal_text_size + j_text_size
            if journal_text_items.any? then
              while total_text_size > MAX_CELL_LENGTH do
                oldest_journal = journal_text_items.shift
                journal_text_size -= oldest_journal.size + 1 #改行文字分も引く
                total_text_size -= oldest_journal.size + 1 #改行文字分も引く 
                if journal_text_items.size == 0 then
                  if j_text_size > MAX_CELL_LENGTH then
                    #最大文字数までは出力する
                    j_text = j_text.slice(0..MAX_CELL_LENGTH - 1)
                    j_text_size = MAX_CELL_LENGTH
                    total_text_size = MAX_CELL_LENGTH
                    journal_text_size = 0
                  end
                  break
                end
              end
            elsif host_journals_size == 1 and total_text_size > MAX_CELL_LENGTH then
              #最大文字数までは出力する
              j_text = j_text.slice(0..MAX_CELL_LENGTH - 1)
              j_text_size = MAX_CELL_LENGTH
            end
            while journal_text_items.any? and journal_text_items.size > MAX_LINE_NUM do
              oldest_journal = journal_text_items.shift
              journal_text_size -= oldest_journal.size
            end
            journal_text_size += j_text_size
            journal_text_items << j_text
          end
          item << journal_text_items.reverse.join("\n")
          item << (host_scsid[:due_date] ? host_scsid[:due_date].strftime('%Y/%m/%d') : '')
          item << ctags.collect { |tag| escape_category_tsv(tag[:keyword]) }.join(',')
          item
        }
        json_tsv_summary_table_progress_download(@hosts[:ids], create_tsv_data, executed_size, 50)
      end
    when 'tsv-package-info' then
      if f_params[:process] == 'init' then
        header = [ 'SIDfm ID', 'ホストID', 'ホスト名', 'パッケージ名' ]
        @hosts[:content_filtered] = @hosts[:content_filtered].where(:sidfmd_host_id => @hosts[:ids]) if @hosts[:ids].any?
        init_tsv_summary_table_progress_download(@hosts[:content_filtered].pluck(:id), header, @hosts[:content_filtered].count)
      else
        create_tsv_data = Proc.new{|id|
          host = @hosts[:content_filtered].find_by(:id => id)
          next unless host
          sidfmd_issue = SidfmdIssue.where(:html_id => @id, :sidfmd_host_id => host[:sidfmd_host_id]).first
          next unless sidfmd_issue
          sidfmd_issue.affect_packages('').each do |pkg|
            item = []
            item << (sidfmd_issue.html_id > 0 ? "#{sidfmd_issue.html_id}" : "KB#{-sidfmd_issue.html_id}")
            item << host[:sidfmd_host_id].to_s
            item << host[:display_name]
            item << pkg[:fullname]
            session[:export_data][:content] << item
          end
          item = nil
        }
        json_tsv_summary_table_progress_download(@hosts[:ids], create_tsv_data, executed_size, 50)
      end
    when 'tsv-affect-products' then
      if f_params[:process] == 'init' then
        header = [ 'SIDfm ID', 'ホストID', 'ホスト名', '製品名' ]
        @hosts[:content_filtered] = @hosts[:content_filtered].where(:sidfmd_host_id => @hosts[:ids]) if @hosts[:ids].any?
        init_tsv_summary_table_progress_download(@hosts[:content_filtered].pluck(:id), header, @hosts[:content_filtered].count)
      else
        create_tsv_data = Proc.new{|id|
          host = @hosts[:content_filtered].find_by(:id => id)
          next unless host
          sidfmd_issue = SidfmdIssue.where(:html_id => @id, :sidfmd_host_id => host[:sidfmd_host_id]).first
          next unless sidfmd_issue
          sidfmd_issue.affect_products.each do |prod|
            item = []
            item << (sidfmd_issue.html_id > 0 ? "#{sidfmd_issue.html_id}" : "KB#{-sidfmd_issue.html_id}")
            item << host[:sidfmd_host_id].to_s
            item << host[:display_name]
            item << prod[:display_name]
            session[:export_data][:content] << item
          end
          sidfmd_issue.affect_lproducts.each do |lprod|
            item = []
            item << (sidfmd_issue.html_id > 0 ? "#{sidfmd_issue.html_id}" : "KB#{-sidfmd_issue.html_id}")
            item << host[:sidfmd_host_id].to_s
            item << host[:display_name]
            item << lprod[:display_name]
            session[:export_data][:content] << item
          end
          item = nil
        }
        json_tsv_summary_table_progress_download(@hosts[:ids], create_tsv_data, executed_size, 50)
      end
    end

  end

  def download
    f_params = params.fetch(:download, {}).permit(:format)
    case f_params[:format]
    when 'json' then
      send_summary_table_json_file("vmlists-vulnerability#{@id}-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json")
    when 'tsv' then
      send_summary_table_tsv_file("vmlists-vulnerability#{@id}-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    when 'tsv-package-info' then
      send_summary_table_tsv_file("vmlists-vulnerability#{@id}-packages-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    when 'tsv-affect-products' then
      send_summary_table_tsv_file("vmlists-vulnerability#{@id}-affect-products-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    end
  end

=begin  
  def download
    pages_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :hostids)
    format = pages_params[:format]
    if ![ 'json', 'tsv', 'tsv-package-info', 'tsv-affect-products' ].include?(format) then
      redirect_to :action => :index, :notice => '不正な入力です。' and return
    end
    @hosts = { :url => request.url }
    @hosts[:category_id] = (pages_params[:category_id] || 0).to_i
    @hosts[:category_name] = '全てのホスト'
    if @hosts[:category_id] > 0 then
      cat = SidfmdCategory.find(@hosts[:category_id])
      @hosts[:category_name] = cat.name if cat
    end
    @hosts[:idx] = (pages_params[:idx] || 'sri').to_sym
    @hosts[:order] = pages_params[:order] || 'hostname_asc'
    @hosts[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    @hosts[:hostids] = []
    begin
      @hosts[:hostids] = JSON.parse(pages_params[:hostids] || '[]')
    rescue JSON::ParserError => e
      logger.error(e)
    end
    @corresp_required = Sidfm.corresp_required
    create_all_content(@hosts[:category_id])
    check_filters
    @hosts[:corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:fixed_corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:fixed_warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:not_fixed_corresp_required_host_size] =
      @hosts[:corresp_required_host_size] - @hosts[:fixed_corresp_required_host_size]
    @hosts[:all_host_size] = @hosts[:all_content].size
    filter_content
    case format
    when 'json' then
      json = {}
      temp_order = @hosts[:order].to_s.split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @hosts[:filters]
      json[:category] = { :id => @hosts[:category_id], :name => @hosts[:category_name], }
      json[:idx] = @hosts[:idx]
      json[:content] = []
      @hosts[:content_filtered].each do |host|
        if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
          next
        end
        item = {}
        item[:host] = { 
          :id => host[:sidfmd_host_id], 
          :hostname => host[:hostname], 
          :ext_name => host[:ext_name],
          :display_name => host[:display_name]
        }
        item[:warning] = host[:warning][@hosts[:idx]]
        item[:status] = SidfmdIssue.status_id_to_s(host[:status_id])
        item[:sri] = SidfmdIssueSri::RISKS[host[:sri]]
        t_cvss = host[:cvss]
        item[:cvss] = (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
        item[:area] =  SidfmdArea.where(:area_id => host[:softek][:area_id]).first.name
        item[:exploitable] = (t_cvss[:label] != :notreg ? host[:exploitable] : nil)
        cats = []
        cats << host[:system_category][:name]
        host[:no_system_categories].each do |cat|
          cats << cat[:name]
        end
        item[:categories] = cats
        item[:os] = ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '')
        item[:journals] =  []
        host[:journals].sort_by{|j| j[:created_on]}.each do |journal|
          j_item = {}
          j_item[:id] = journal[:id]
          j_item[:created_on]  = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : '---'
          user = User.find(journal[:user_id])
          j_item[:user] = { :id => journal[:user_id], :name => user.anonymous? ? '削除済みユーザ' : view_context.display_user_name(user) }
          j_item[:notes] = Journal.find(journal[:id]).notes
          j_item[:details] = []
          JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
            jd_item = {}
            jd_item[:id] = detail[:id]
            case detail[:prop_key]
            when 'status_id' then
              jd_item[:action] = 'ステータス変更'
              jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
              jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
            when 'due_date' then
              jd_item[:action] = '対策予定日の変更'
              jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
              jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
            end
            j_item[:details] << jd_item
          end
          item[:journals] << j_item
        end
        item[:journals].reverse!
        item[:due_date] = host[:due_date]
        apps = []
        pkgs = []
        happs = []
        hlapps = []
        sidfmd_host = SidfmdHost.find(host[:sidfmd_host_id])
        if sidfmd_host then
          queue = sidfmd_host.get_last_queue
          if queue then
            pkgs = queue.sidfmd_queue_pkgs.order(:pkg).collect {|x| x.pkg }
            apps = queue.sidfmd_queue_apps.order(:name,:version).collect {|x| 
              { :name => x.name, :version => x.version, :path => x.path }
            }
          end
          happs = sidfmd_host.sidfmd_host_apps.order(:product_id,:version_id).collect {|x| x.get_name }.compact
          hlapps = sidfmd_host.sidfmd_host_lapps.order(:sidfmd_lproduct_id, :sidfmd_lversion_id).collect {|x| x.get_name }.compact
        end
        item[:packages] = pkgs
        item[:products] = apps
        item[:manual_products] = happs
        item[:local_products] = hlapps

        item[:affect_packages] = []
        item[:affect_products] = []

        item[:issue] = { :id => host[:sidfmd_issue_id] }
        issue = SidfmdIssue.find(host[:sidfmd_issue_id])
        if issue then
          item[:affect_packages] = issue.affect_packages('')
          item[:affect_products] = issue.affect_products + issue.affect_lproducts
          item[:issue][:title] = issue.title
          item[:issue][:html_id] = issue.html_id > 0 ? "#{issue.html_id}" : "KB#{-issue.html_id}"
        end
        json[:content] << item
      end
      send_data(json.to_json, :filename => "vmlists-vulnerability#{@id}-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
    when 'tsv' then
      convert_chars = ->(text){
        text.gsub('"', '”').gsub("'", "‘").gsub(',', '，').gsub("\n", ' ')
      }
      header = [ 'ホストID', '要対応', '対応状況', 'SRI', 'CVSS', '影響範囲', '攻撃コード', 'ホスト名', '表示名', 'カテゴリ', 'OS名', '履歴数', '履歴(日時,ユーザ名,メールアドレス,操作,操作内容,コメント)', '対策予定日' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          item.clear
          item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
          item << host[:warning][@hosts[:idx]].to_s.encode('UTF-16LE')
          item << SidfmdIssue.status_id_to_s(host[:status_id]).encode('UTF-16LE')
          item << SidfmdIssueSri::RISKS[host[:sri]].encode('UTF-16LE')
          t_cvss = host[:cvss]
          item << (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => host[:softek][:area_id]).first.name.encode('UTF-16LE')
          item << host[:exploitable].to_s.encode('UTF-16LE')
          item << host[:hostname].encode('UTF-16LE')
          item << host[:display_name].encode('UTF-16LE')
          cats = []
          cats << escape_category_tsv(host[:system_category][:name]).encode('UTF-16LE')
          host[:no_system_categories].each do |cat|
            cats << escape_category_tsv(cat[:name]).encode('UTF-16LE')
          end
          item << cats.join(','.encode('UTF-16LE'))
          item << ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '').encode('UTF-16LE')
          item << host[:journals].size.to_s.encode('UTF-16LE')
          count_journal_lines = 0
          journal_text_size = 0
          journal_text_items = []
          host[:journals].sort_by{|j| j[:created_on]}.each_with_index do |journal, idx|
            j_text = ''
            if journal[:created_on] then
              j_text += journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') + ','
            end
            user = User.find(journal[:user_id])
            j_text += user.anonymous? ?  '削除済みユーザ,' : convert_chars.(view_context.display_user_name(user)) + ','
            j_text += user.mail ? "#{user.mail}," : ','
            detail_flag = false
            JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
              case detail[:prop_key]
              when 'status_id' then
                detail_flag = true
                j_text += 'ステータス変更' + ','
                j_text += "[#{SidfmdIssue.status_id_to_s(detail[:old_value].to_i)}]から[#{SidfmdIssue.status_id_to_s(detail[:value].to_i)}]へ,"
              when 'due_date' then
                detail_flag = true
                j_text += '対策予定日の変更' + ','
                j_text += "[#{SidfmdIssue.due_date_to_s(detail[:old_value])}]から[#{SidfmdIssue.due_date_to_s(detail[:value])}]へ,"
              end
            end
            if !detail_flag then
              j_text += ',,'
            end
            j_text += convert_chars.(Journal.find(journal[:id]).notes)
            j_text_size = j_text.size
            if host[:journals].size != 1 then
              if idx != 0 then
                j_text_size += 1 #改行文字をカウント
              end
            end
            total_text_size = journal_text_size + j_text_size
            if journal_text_items.any? then
              while total_text_size > MAX_CELL_LENGTH do 
                oldest_journal = journal_text_items.shift
                journal_text_size -= oldest_journal.size + 1 #改行文字分も引く
                total_text_size -= oldest_journal.size + 1 #改行文字分も引く 
                if journal_text_items.size == 0 then 
                  if j_text_size > MAX_CELL_LENGTH then
                    #最大文字数までは出力する
                    j_text = j_text.slice(0..MAX_CELL_LENGTH - 1)
                    j_text_size = MAX_CELL_LENGTH
                    total_text_size = MAX_CELL_LENGTH
                    journal_text_size = 0
                  end
                  break
                end
              end
            elsif host[:journals].size == 1 and total_text_size > MAX_CELL_LENGTH then
              #最大文字数までは出力する
              j_text = j_text.slice(0..MAX_CELL_LENGTH - 1)
              j_text_size = MAX_CELL_LENGTH
            end
            while journal_text_items.any? and journal_text_items.size > MAX_LINE_NUM do
              oldest_journal = journal_text_items.shift
              journal_text_size -= oldest_journal.size
            end
            journal_text_size += j_text_size
            journal_text_items << j_text
          end
          item << journal_text_items.reverse.join("\n").encode('UTF-16LE')
          item << (host[:due_date] ? host[:due_date] : '').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlists-vulnerability#{@id}-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
    when 'tsv-package-info' then
      header = [ 'SIDfm ID', 'ホストID', 'ホスト名', 'パッケージ名' ]
      header = header.map {|x| x.encode('UTF-16LE') } 
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          sidfmd_issue = SidfmdIssue.find_by(:html_id => @id, :sidfmd_host_id => host[:sidfmd_host_id])
          if sidfmd_issue then
            sidfmd_issue.affect_packages('').each do |pkg|
              item.clear
              item << @id.to_s.encode('UTF-16LE')
              item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
              item << host[:display_name].encode('UTF-16LE')
              item << pkg[:fullname].encode('UTF-16LE')
              o << item
            end
          end
        end
      end
      send_data(csv, :filename => "vmlists-vulnerability#{@id}-packages-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
    when 'tsv-affect-products' then
      header = [ 'SIDfm ID', 'ホストID', 'ホスト名', '製品名' ] 
      header = header.map {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          sidfmd_issue = SidfmdIssue.find_by(:html_id => @id, :sidfmd_host_id => host[:sidfmd_host_id])
          if sidfmd_issue then
            sidfmd_issue.affect_products.each do |prod|
              item.clear
              item << @id.to_s.encode('UTF-16LE')
              item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
              item << host[:display_name].encode('UTF-16LE')
              item << prod[:display_name] .encode('UTF-16LE')
              o << item
            end
            sidfmd_issue.affect_lproducts.each do |lprod|
              item.clear
              item << @id.to_s.encode('UTF-16LE')
              item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
              item << host[:display_name].encode('UTF-16LE')
              item << lprod[:display_name] .encode('UTF-16LE')
              o << item 
            end
          end
        end
      end
      send_data(csv, :filename => "vmlists-vulnerability#{@id}-affect-products-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
    end

    return
    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end
=end
  
  def send_mail
    filtered_params = params.permit(:recipients, :subject, :message)
  begin
    mail = {}
    mail[:from] = User.current.mail
    email_with_name = []
    recipients = JSON.parse(filtered_params[:recipients])
    recipients.each do |address, name|
      email_name = %("#{name}" <#{address}>)
      email_with_name << email_name
    end
    #mail[:to] = email_with_name
    mail[:subject] = filtered_params[:subject]
    mail[:message] = filtered_params[:message]
    send_mail_num = SidfmdMailer::SEND_MAIL_NUM
    time_out_second = SidfmdMailer::SEND_MAIL_TIMEOUT_SECOND
    while (send_mail_targets = email_with_name.shift(send_mail_num)).size > 0 do
      threads = send_mail_targets.map{|fullname_and_address|
        Thread.new(mail.deep_dup, fullname_and_address){|cp_mail, data|
          ret = {}
          cp_mail[:to] = data
          begin
            SidfmdMailer.send_mail(cp_mail).deliver_now
          rescue Exception => e
            ret[:exception] = e
          end
          ret
        }
      }
      threads.each do |result|
        result = result.join(time_out_second)
        if result then
          result = result.value
          if result[:exception] then logger.error("failed send mail. #{result[:exception]}") end
        else
          logger.error("time-out of sending a mail")
        end
      end
    end
    #SidfmdMailer.send_mail(mail).deliver_now
  rescue Exception => e
    log_exception(e)
    render :layout => false, :json => { :err_message => 'メールの送信に失敗しました。' } and return
    #redirect_to :action => :index, :notice => e and return
  end
    render :layout => false, :json => { :err_message => '' }
  end

=begin
  def api_get_display_summary_rows
    GC.start
    f_params = params.permit(:ids, :category_id, :idx, :checked, :filters)
    category_id = (f_params[:category_id] || 0).to_i
    host_ids = JSON.parse(f_params[:ids])
    filters = JSON.parse(f_params[:filters], :symbolize_names => true)
    checked = JSON.parse(f_params[:checked]).map(&:to_i)
    @hosts = {}
    create_host_view_content(category_id)
    @target_scsids = SidfmdVvScsid.where(:sidfmd_host_id => @hosts[:all_content].select(:sidfmd_host_id))
                                  .where(:sidfmd_vul_view_id => @vul_view.id)
    idx = f_params[:idx].to_sym
    corresp_required = Sidfm.corresp_required
    get_summary_prior_category
    user_id = User.current.id
    ret = {}
    @hosts[:all_content].where(:sidfmd_host_id => host_ids).each do |host_view|
      ret[host_view.sidfmd_host_id] = create_summary_host_html(
        host_view,
        {
          :category_id => category_id,
          :idx => idx,
          :corresp_required => corresp_required,
          :checked =>  checked,
          :summary_prior_category => @summary_prior_category,
          :scsid => @target_scsids.find_by(:sidfmd_host_id => host_view.sidfmd_host_id),
          :user_id => user_id 
        }
      )
    end

    render :layout => false, :json => ret
  end
=end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:per_page, :history_filter_host)
    parse_params
    @hosts[:per_page] = f_params[:per_page].to_i
    history_filter_host = JSON.parse((f_params[:history_filter_host] || '{}'), symbolize_names: true)
    get_summary_prior_category
    @corresp_required = Sidfm.corresp_required
    create_host_view_content(@hosts[:category_id])
    check_filters
    is_history_filter_host = history_filter_host.size > 0
    @hosts[:content] = is_history_filter_host ?
      @hosts[:content_filtered].where(:sidfmd_host_id => history_filter_host[:sidfmd_host_id]).limit(1) :
      @hosts[:content_filtered].limit(@hosts[:per_page]).offset(@hosts[:offset])
    user_id = User.current.id
    create_host_ctags(@hosts[:content].pluck(:id))
    ret = {
      :rows => @hosts[:content].map{|host_view|
        create_summary_host_html(
          host_view,
          {
            :category_id => @hosts[:category_id],
            :idx => @hosts[:idx],
            :corresp_required => @corresp_required,
            :checked =>  @hosts[:checked],
            :summary_prior_category => @summary_prior_category,
            :scsid => @target_scsids.find_by(:sidfmd_host_id => host_view.sidfmd_host_id),
            :user_id => user_id
          }
        )
      },
      :total => (is_history_filter_host ? @hosts[:content] : @hosts[:content_filtered]).count,
      :all_ids => (is_history_filter_host ? @hosts[:content] : @hosts[:content_filtered]).pluck(:sidfmd_host_id),
      :sidfm_roles => @sidfm_roles,
      :user_cats => @user_category_ids
    }
    render :layout => false, :json => ret
  end

  def api_establish_filters_consistency
    f_params = params.permit(:filters, :user_host_osnames)
    filters = JSON.parse(f_params[:filters], :symbolize_names => true)
    filters[:categories][:ids] = filters[:categories][:ids].map(&:to_i) if filters[:categories] && filters[:categories][:ids]
    user_host_osnames = JSON.parse(f_params[:user_host_osnames])
    @hosts = {}
    create_host_view_content
    init_filter_keys
    filters = establish_filters_consistency(filters, @filter_keys)
    filters = filter_filters(
      :filters => filters,
      :user_host_osnames => user_host_osnames,
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids
    )
    render :layout => false, :json => filters
  end

end

